@@ -5,6 +5,7 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE OverloadedLabels #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TypeInType #-}
 {-# LANGUAGE TypeOperators #-}
@@ -43,16 +44,16 @@ import Control.Monad
 import Control.Monad.Catch (MonadThrow)
 import Control.Monad.Except
 import Control.Monad.IO.Class
+import Control.Monad.Trans.Class
 #if __GLASGOW_HASKELL__ == 806
 import Control.Monad.Fail
 #endif
-import Control.Monad.Trans.Class
 import Control.Monad.Trans.Reader (ReaderT, runReaderT)
 import qualified Control.Monad.Trans.Reader as Reader (ask)
 import Control.Monad.Trans.State (StateT, runStateT, execState)
 import qualified Control.Monad.Trans.State as State
 import qualified Data.ByteString.Lazy.Char8 as B
-import Data.Aeson hiding (Error)
+import Data.Aeson hiding (Error, Null)
 import Data.Aeson.Encode.Pretty
 import Data.Conduit as Conduit
 import Data.Conduit.Parser as Parser
@@ -63,13 +64,11 @@ import qualified Data.Map.Strict as Map
 import qualified Data.Set as Set
 import qualified Data.Text as T
 import qualified Data.Text.IO as T
-import qualified Data.HashMap.Strict as HashMap
 import Data.Maybe
 import Data.Function
-import Language.LSP.Types.Capabilities
-import Language.LSP.Types
-import Language.LSP.Types.Lens
-import qualified Language.LSP.Types.Lens as LSP
+import Language.LSP.Protocol.Types as LSP
+import qualified Language.LSP.Protocol.Lens as L
+import Language.LSP.Protocol.Message as LSP
 import Language.LSP.VFS
 import Language.LSP.Test.Compat
 import Language.LSP.Test.Decoding
@@ -84,6 +83,7 @@ import System.Process (waitForProcess)
 import System.Timeout ( timeout )
 import Data.IORef
 import Colog.Core (LogAction (..), WithSeverity (..), Severity (..))
+import Data.Row
 
 -- | A session representing one instance of launching and connecting to a server.
 --
@@ -142,7 +142,7 @@ data SessionContext = SessionContext
   -- Keep curTimeoutId in SessionContext, as its tied to messageChan
   , curTimeoutId :: IORef Int -- ^ The current timeout we are waiting on
   , requestMap :: MVar RequestMap
-  , initRsp :: MVar (ResponseMessage Initialize)
+  , initRsp :: MVar (TResponseMessage Method_Initialize)
   , config :: SessionConfig
   , sessionCapabilities :: ClientCapabilities
   }
@@ -231,9 +231,9 @@ runSessionMonad context state (Session session) = runReaderT (runStateT conduit
         yield msg
       chanSource
 
-    isLogNotification (ServerMessage (FromServerMess SWindowShowMessage _)) = True
-    isLogNotification (ServerMessage (FromServerMess SWindowLogMessage _)) = True
-    isLogNotification (ServerMessage (FromServerMess SWindowShowDocument _)) = True
+    isLogNotification (ServerMessage (FromServerMess SMethod_WindowShowMessage _)) = True
+    isLogNotification (ServerMessage (FromServerMess SMethod_WindowLogMessage _)) = True
+    isLogNotification (ServerMessage (FromServerMess SMethod_WindowShowDocument _)) = True
     isLogNotification _ = False
 
     watchdog :: ConduitM SessionMessage FromServerMessage (StateT SessionState (ReaderT SessionContext IO)) ()
@@ -307,64 +307,67 @@ updateStateC = awaitForever $ \msg -> do
   yield msg
   where
     respond :: (MonadIO m, HasReader SessionContext m) => FromServerMessage -> m ()
-    respond (FromServerMess SWindowWorkDoneProgressCreate req) =
-      sendMessage $ ResponseMessage "2.0" (Just $ req ^. LSP.id) (Right Empty)
-    respond (FromServerMess SWorkspaceApplyEdit r) = do
-      sendMessage $ ResponseMessage "2.0" (Just $ r ^. LSP.id) (Right $ ApplyWorkspaceEditResponseBody True Nothing Nothing)
+    respond (FromServerMess SMethod_WindowWorkDoneProgressCreate req) =
+      sendMessage $ TResponseMessage "2.0" (Just $ req ^. L.id) (Right Null)
+    respond (FromServerMess SMethod_WorkspaceApplyEdit r) = do
+      sendMessage $ TResponseMessage "2.0" (Just $ r ^. L.id) (Right $ ApplyWorkspaceEditResult True Nothing Nothing)
     respond _ = pure ()
 
 
 -- extract Uri out from DocumentChange
 -- didn't put this in `lsp-types` because TH was getting in the way
 documentChangeUri :: DocumentChange -> Uri
-documentChangeUri (InL x) = x ^. textDocument . uri
-documentChangeUri (InR (InL x)) = x ^. uri
-documentChangeUri (InR (InR (InL x))) = x ^. oldUri
-documentChangeUri (InR (InR (InR x))) = x ^. uri
+documentChangeUri (InL x) = x ^. L.textDocument . L.uri
+documentChangeUri (InR (InL x)) = x ^. L.uri
+documentChangeUri (InR (InR (InL x))) = x ^. L.oldUri
+documentChangeUri (InR (InR (InR x))) = x ^. L.uri
 
 updateState :: (MonadIO m, HasReader SessionContext m, HasState SessionState m)
             => FromServerMessage -> m ()
-updateState (FromServerMess SProgress req) = case req ^. params . value of
-  Begin _ ->
-    modify $ \s -> s { curProgressSessions = Set.insert (req ^. params . token) $ curProgressSessions s }
-  End _ ->
-    modify $ \s -> s { curProgressSessions = Set.delete (req ^. params . token) $ curProgressSessions s }
+updateState (FromServerMess SMethod_Progress req) = case req ^. L.params . L.value of
+  v | Just _ <- v ^? _workDoneProgressBegin ->
+    modify $ \s -> s { curProgressSessions = Set.insert (req ^. L.params . L.token) $ curProgressSessions s }
+  v | Just _ <- v ^? _workDoneProgressEnd ->
+    modify $ \s -> s { curProgressSessions = Set.delete (req ^. L.params . L.token) $ curProgressSessions s }
   _ -> pure ()
 
 -- Keep track of dynamic capability registration
-updateState (FromServerMess SClientRegisterCapability req) = do
-  let List newRegs = (\sr@(SomeRegistration r) -> (r ^. LSP.id, sr)) <$> req ^. params . registrations
+updateState (FromServerMess SMethod_ClientRegisterCapability req) = do
+  let
+    regs :: [SomeRegistration]
+    regs = req ^.. L.params . L.registrations . traversed . to toSomeRegistration . _Just
+  let newRegs = (\sr@(SomeRegistration r) -> (r ^. L.id, sr)) <$> regs
   modify $ \s ->
     s { curDynCaps = Map.union (Map.fromList newRegs) (curDynCaps s) }
 
-updateState (FromServerMess SClientUnregisterCapability req) = do
-  let List unRegs = (^. LSP.id) <$> req ^. params . unregisterations
+updateState (FromServerMess SMethod_ClientUnregisterCapability req) = do
+  let unRegs = (^. L.id) <$> req ^. L.params . L.unregisterations
   modify $ \s ->
     let newCurDynCaps = foldr' Map.delete (curDynCaps s) unRegs
     in s { curDynCaps = newCurDynCaps }
 
-updateState (FromServerMess STextDocumentPublishDiagnostics n) = do
-  let List diags = n ^. params . diagnostics
-      doc = n ^. params . uri
+updateState (FromServerMess SMethod_TextDocumentPublishDiagnostics n) = do
+  let diags = n ^. L.params . L.diagnostics
+      doc = n ^. L.params . L.uri
   modify $ \s ->
     let newDiags = Map.insert (toNormalizedUri doc) diags (curDiagnostics s)
       in s { curDiagnostics = newDiags }
 
-updateState (FromServerMess SWorkspaceApplyEdit r) = do
+updateState (FromServerMess SMethod_WorkspaceApplyEdit r) = do
 
   -- First, prefer the versioned documentChanges field
-  allChangeParams <- case r ^. params . edit . documentChanges of
-    Just (List cs) -> do
+  allChangeParams <- case r ^. L.params . L.edit . L.documentChanges of
+    Just (cs) -> do
       mapM_ (checkIfNeedsOpened . documentChangeUri) cs
       -- replace the user provided version numbers with the VFS ones + 1
       -- (technically we should check that the user versions match the VFS ones)
-      cs' <- traverseOf (traverse . _InL . textDocument) bumpNewestVersion cs
+      cs' <- traverseOf (traverse . _L . L.textDocument . _versionedTextDocumentIdentifier) bumpNewestVersion cs
       return $ mapMaybe getParamsFromDocumentChange cs'
     -- Then fall back to the changes field
-    Nothing -> case r ^. params . edit . changes of
+    Nothing -> case r ^. L.params . L.edit . L.changes of
       Just cs -> do
-        mapM_ checkIfNeedsOpened (HashMap.keys cs)
-        concat <$> mapM (uncurry getChangeParams) (HashMap.toList cs)
+        mapM_ checkIfNeedsOpened (Map.keys cs)
+        concat <$> mapM (uncurry getChangeParams) (Map.toList cs)
       Nothing ->
         error "WorkspaceEdit contains neither documentChanges nor changes!"
 
@@ -372,20 +375,20 @@ updateState (FromServerMess SWorkspaceApplyEdit r) = do
     let newVFS = flip execState (vfs s) $ changeFromServerVFS logger r
     return $ s { vfs = newVFS }
 
-  let groupedParams = groupBy (\a b -> a ^. textDocument == b ^. textDocument) allChangeParams
+  let groupedParams = groupBy (\a b -> a ^. L.textDocument == b ^. L.textDocument) allChangeParams
       mergedParams = map mergeParams groupedParams
 
   -- TODO: Don't do this when replaying a session
-  forM_ mergedParams (sendMessage . NotificationMessage "2.0" STextDocumentDidChange)
+  forM_ mergedParams (sendMessage . TNotificationMessage "2.0" SMethod_TextDocumentDidChange)
 
   -- Update VFS to new document versions
-  let sortedVersions = map (sortBy (compare `on` (^. textDocument . version))) groupedParams
-      latestVersions = map ((^. textDocument) . last) sortedVersions
+  let sortedVersions = map (sortBy (compare `on` (^. L.textDocument . L.version))) groupedParams
+      latestVersions = map ((^. L.textDocument) . last) sortedVersions
 
   forM_ latestVersions $ \(VersionedTextDocumentIdentifier uri v) ->
     modify $ \s ->
       let oldVFS = vfs s
-          update (VirtualFile oldV file_ver t) = VirtualFile (fromMaybe oldV v) (file_ver +1) t
+          update (VirtualFile _ file_ver t) = VirtualFile v (file_ver +1) t
           newVFS = oldVFS & vfsMap . ix (toNormalizedUri uri) %~ update
       in s { vfs = newVFS }
 
@@ -399,23 +402,24 @@ updateState (FromServerMess SWorkspaceApplyEdit r) = do
             let fp = fromJust $ uriToFilePath uri
             contents <- liftIO $ T.readFile fp
             let item = TextDocumentItem (filePathToUri fp) "" 0 contents
-                msg = NotificationMessage "2.0" STextDocumentDidOpen (DidOpenTextDocumentParams item)
+                msg = TNotificationMessage "2.0" SMethod_TextDocumentDidOpen (DidOpenTextDocumentParams item)
             sendMessage msg
 
             modifyM $ \s -> do
               let newVFS = flip execState (vfs s) $ openVFS logger msg
               return $ s { vfs = newVFS }
 
-        getParamsFromTextDocumentEdit :: TextDocumentEdit -> DidChangeTextDocumentParams
-        getParamsFromTextDocumentEdit (TextDocumentEdit docId (List edits)) = do
-          DidChangeTextDocumentParams docId (List $ map editToChangeEvent edits)
+        getParamsFromTextDocumentEdit :: TextDocumentEdit -> Maybe DidChangeTextDocumentParams
+        getParamsFromTextDocumentEdit (TextDocumentEdit docId edits) =
+          DidChangeTextDocumentParams <$> docId ^? _versionedTextDocumentIdentifier <*> pure (map editToChangeEvent edits)
 
+        -- TODO: move somewhere reusable
         editToChangeEvent :: TextEdit |? AnnotatedTextEdit -> TextDocumentContentChangeEvent
-        editToChangeEvent (InR e) = TextDocumentContentChangeEvent (Just $ e ^. range) Nothing (e ^. newText)
-        editToChangeEvent (InL e) = TextDocumentContentChangeEvent (Just $ e ^. range) Nothing (e ^. newText)
+        editToChangeEvent (InR e) = TextDocumentContentChangeEvent $ InL $ #range .== (e ^. L.range) .+ #rangeLength .== Nothing .+ #text .== (e ^. L.newText)
+        editToChangeEvent (InL e) = TextDocumentContentChangeEvent $ InL $ #range .== (e ^. L.range) .+ #rangeLength .== Nothing .+ #text .== (e ^. L.newText)
 
         getParamsFromDocumentChange :: DocumentChange -> Maybe DidChangeTextDocumentParams
-        getParamsFromDocumentChange (InL textDocumentEdit) = Just $ getParamsFromTextDocumentEdit textDocumentEdit
+        getParamsFromDocumentChange (InL textDocumentEdit) = getParamsFromTextDocumentEdit textDocumentEdit
         getParamsFromDocumentChange _ = Nothing
 
         bumpNewestVersion (VersionedTextDocumentIdentifier uri _) =
@@ -426,18 +430,19 @@ updateState (FromServerMess SWorkspaceApplyEdit r) = do
         textDocumentVersions uri = do
           vfs <- vfs <$> get
           let curVer = fromMaybe 0 $ vfs ^? vfsMap . ix (toNormalizedUri uri) . lsp_version
-          pure $ map (VersionedTextDocumentIdentifier uri . Just) [curVer + 1..]
+          pure $ map (VersionedTextDocumentIdentifier uri) [curVer + 1..]
 
         textDocumentEdits uri edits = do
           vers <- textDocumentVersions uri
-          pure $ map (\(v, e) -> TextDocumentEdit v (List [InL e])) $ zip vers edits
+          pure $ map (\(v, e) -> TextDocumentEdit (review _versionedTextDocumentIdentifier v) [InL e]) $ zip vers edits
 
-        getChangeParams uri (List edits) = do
-          map <$> pure getParamsFromTextDocumentEdit <*> textDocumentEdits uri (reverse edits)
+        getChangeParams uri edits = do
+          edits <- textDocumentEdits uri (reverse edits)
+          pure $ catMaybes $ map getParamsFromTextDocumentEdit edits
 
         mergeParams :: [DidChangeTextDocumentParams] -> DidChangeTextDocumentParams
-        mergeParams params = let events = concat (toList (map (toList . (^. contentChanges)) params))
-                              in DidChangeTextDocumentParams (head params ^. textDocument) (List events)
+        mergeParams params = let events = concat (toList (map (toList . (^. L.contentChanges)) params))
+                              in DidChangeTextDocumentParams (head params ^. L.textDocument) events
 updateState _ = return ()
 
 sendMessage :: (MonadIO m, HasReader SessionContext m, ToJSON a) => a -> m ()
